# *
# *	Copyright (c) 2000 Alberto Reggiori / <alberto.reggiori@jrc.it>
# *	ISIS/RIT, Joint Research Center Ispra (I)
# *
# * NOTICE
# *
# * This product is distributed under a BSD/ASF like license as described in the 'LICENSE'
# * file you should have received together with this source code. If you did not get a
# * a copy of such a license agreement you can pick up one at:
# *
# *     http://xml.jrc.it/RDFStore/LICENSE
# *
# *
package XML::Parser::PenRDF;
{
	use strict;

	use vars qw{$VERSION %Built_In_Styles};
	use Carp qw(carp croak cluck confess);
	use URI;
	use URI::file;
	use URI::Escape;
	require "chat2.pl"; # :)

BEGIN
{
	require XML::Parser::Expat;
    	$VERSION = '0.1';
    	croak "XML::Parser::Expat.pm version 2 or higher is needed"
		unless $XML::Parser::Expat::VERSION =~ /^2\./;
}

$XML::Parser::PenRDF::RDF_SYNTAX_NS="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
$XML::Parser::PenRDF::RDF_SCHEMA_NS="http://www.w3.org/TR/1999/PR-rdf-schema-19990303#";
$XML::Parser::PenRDF::XMLSCHEMA="xml";
$XML::Parser::PenRDF::XML_space=$XML::Parser::PenRDF::XMLSCHEMA."space";
$XML::Parser::PenRDF::XML_space_preserve="preserve";
$XML::Parser::PenRDF::XMLNS="xmlns";
$XML::Parser::PenRDF::RDFMS_parseType = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "parseType";
$XML::Parser::PenRDF::RDFMS_type = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "type";
$XML::Parser::PenRDF::RDFMS_about = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "about";
$XML::Parser::PenRDF::RDFMS_bagID = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "bagID";
$XML::Parser::PenRDF::RDFMS_resource = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "resource";
$XML::Parser::PenRDF::RDFMS_aboutEach = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "aboutEach";
$XML::Parser::PenRDF::RDFMS_aboutEachPrefix = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "aboutEachPrefix";
$XML::Parser::PenRDF::RDFMS_ID = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "ID";
$XML::Parser::PenRDF::RDFMS_RDF = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "RDF";
$XML::Parser::PenRDF::RDFMS_Description = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "Description";
$XML::Parser::PenRDF::RDFMS_Seq = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "Seq";
$XML::Parser::PenRDF::RDFMS_Alt = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "Alt";
$XML::Parser::PenRDF::RDFMS_Bag = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "Bag";
$XML::Parser::PenRDF::RDFMS_predicate = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "predicate";
$XML::Parser::PenRDF::RDFMS_subject = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "subject";
$XML::Parser::PenRDF::RDFMS_object = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "object";
$XML::Parser::PenRDF::RDFMS_Statement = $XML::Parser::PenRDF::RDF_SYNTAX_NS . "Statement";

sub new
{
    	my ($class, %args) = @_;
    	my $style = $args{Style};

	$args{_HNDL_TYPES} = {%XML::Parser::Expat::Handler_Setters};
	$args{_HNDL_TYPES}->{Init} = 1;
	$args{_HNDL_TYPES}->{Assert} = 1;
	$args{_HNDL_TYPES}->{Start_XML_Literal} = 1;
	$args{_HNDL_TYPES}->{Stop_XML_Literal} = 1;
	$args{_HNDL_TYPES}->{Char_Literal} = 1;
	$args{_HNDL_TYPES}->{Final} = 1;

    	my $handlers = $args{Handlers};
    	if (defined($style))
	{
		my $stylepkg = $style;
		if ($stylepkg !~ /::/)
		{
	    		$stylepkg = "\u$style";
	    		croak "Undefined style: $style" 
				unless defined($Built_In_Styles{$stylepkg});
	    		$stylepkg = 'XML::Parser::PenRDF::' . $stylepkg;
		}
		unless (defined($handlers))
		{
			$args{Handlers} = $handlers = {};
		}

		my $htype;
		foreach $htype (keys %{$args{_HNDL_TYPES}})
		{
	    		# Handlers explicity given override
	    		# handlers from the Style package
	    		unless (defined($handlers->{$htype}))
			{
				# A handler in the style package must either have
				# exactly the right case as the type name or a
				# completely lower case version of it.
				my $hname = "${stylepkg}::$htype";
				if (defined(&$hname))
				{
		    			$handlers->{$htype} = \&$hname;
		    			next;
				}
				$hname = "${stylepkg}::\L$htype";
				if (defined(&$hname))
				{
		    			$handlers->{$htype} = \&$hname;
		    			next;
				}
	    		}
		}
	}
    	$args{Pkg} ||= caller;

    	bless \%args, $class;
}

sub setHandlers
{
	my ($class, @handler_pairs) = @_;

	croak("Uneven number of arguments to setHandlers method") 
		if (int(@handler_pairs) & 1);

	my @ret;
	while (@handler_pairs)
	{
		my $type = shift @handler_pairs;
		my $handler = shift @handler_pairs;
		unless (defined($class->{_HNDL_TYPES}->{$type})) {
			my @types = sort keys %{$class->{_HNDL_TYPES}};
	    		croak("Unknown Parser handler type: $type\n Valid types are : @types");
		}
		push(@ret, $type, $class->{Handlers}->{$type});
		$class->{Handlers}->{$type} = $handler;
    	};
	return @ret;
}
sub setSource {
  	my ($file_or_uri)=@_;
	$file_or_uri .= '#'
		if(	(defined $file_or_uri) &&
			(!($file_or_uri =~ /#$/)) ||
			($file_or_uri =~ /\/$/) ||
			($file_or_uri =~ /:$/)
			);
	return $file_or_uri;
};

sub parse
{
	my ($class,$arg,$file_or_uri) = @_;

	my @expat_options = ( Namespaces => 1 ); #RDF needs them :)
	my ($key, $val);
	while (($key, $val) = each %{$class})
      	{
		push(@expat_options, $key, $val) 
			unless exists $class->{Non_Expat_Options}->{$key};
      	}

	#Run Expat
    	my $first = new XML::Parser::Expat(@expat_options, @_);

	$first->{PenRDF} = {};

	#from libwww & SiRPAC
  	$first->{PenRDF}->{elementStack} = [];
  	$first->{PenRDF}->{root}='';
  	$first->{PenRDF}->{EXPECT_Element}='';
  	$first->{PenRDF}->{iReificationCounter}=0;
  	$first->{PenRDF}->{sSource}= ( (exists $class->{Source}) && (defined $class->{Source}) ) ?
					setSource($class->{Source}) : setSource($file_or_uri);

	# The walk-through of RDF schemas requires two lists
	# shared by all SiRPAC instances
	# 1. s_vNStodo - list of all namespaces SiRPAC should still vist
	# 2. s_vNSdone - list of all namespaces SiRPAC has gone through
  	$first->{PenRDF}->{vNStodo}=[];
  	$first->{PenRDF}->{vNSdone}=[];

	# The following two variables may be changed on the fly
	# to change the behaviour of the parser
	#
	# createBags method allows one to determine whether SiRPAC
	# produces Bag instances for each Description block.
	# The default setting is to generate them. - to be checked......
  	$first->{PenRDF}->{bCreateBags}=0;
	#
	# Set whether parser recursively fetches and parses
	# every RDF schema it finds in the namespace declarations
  	$first->{PenRDF}->{bFetchSchemas}=0;

	# The following flag indicates whether the XML markup
	# should be stored into a string as a literal value for RDF
  	$first->{PenRDF}->{parseElementStack} = [];
  	$first->{PenRDF}->{parseTypeStack} = [];
  	$first->{PenRDF}->{scanMode} = 'SKIPPING';
	$first->{PenRDF}->{sLiteral} = '';
	croak "Missing NodeFactory"
		unless( (defined $class->{NodeFactory}) && ($class->{NodeFactory}->isa("RDF::NodeFactory")) );
  	$first->{PenRDF}->{nodeFactory} = $class->{NodeFactory};

    	$first->setHandlers(	Start => \&RDFXML_StartElementHandler,
				End => \&RDFXML_EndElementHandler,
				Char => \&RDFXML_CharacterDataHandler );
	$first->{Extra_Options} = $class->{Extra_Options};

	$class->{handlers} = {};
    	%{$class->{handlers}} = %{$class->{Handlers}} 
		if( (defined $class->{Handlers}) && (ref($class->{Handlers}) =~ /HASH/) );

    	my $init = delete $class->{handlers}->{Init};
    	my $final = delete $class->{handlers}->{Final};

	#Trigger 'Init' event
    	&$init($first) 
		if defined($init);

    	my $result = $first->parse($arg);

	$first->{parser_parameters} = [];
	@{$first->{parser_parameters}} = (@expat_options,@_);

    	if (defined($final))
	{
		#Trigger 'Final' event
		$result = &$final($first);
	}

    	return $result;
};

sub parsestring
{
	my $class = shift;
	return $class->parse(@_);
}

sub parsefile
{
	my $class = shift;
	my $file = shift;

	if( (defined $file) && ($file ne '') ) {
		my $ret;
		my $file_uri;
		#anyway it does not work for fullpathname files unless use 'file:' :(
		if(-e $file) {
                        $file_uri= URI::file->new_abs($file);
                } elsif(!($file =~ /^[\w]+:/)) {
                        $file_uri= URI::file->new_abs($file,URI::file->cwd);
                } else {
                        $file_uri= URI->new($file);
                };
		if (	(defined $file_uri->scheme)	&&
			($file_uri->scheme ne 'file') ) {
    			my $uri = new URI($file);
  			my $wget_handle = wget($uri);
			if(defined $wget_handle) {
				eval {
					$ret = $class->parse($wget_handle, $uri);
    				};
    				my $err = $@;
    				croak $err 	
					if $err;
                        } else {
				croak "Cannot fetch '$uri'";
			};
    		} else {
			my $file= $file_uri->file;

			croak "Unable to open input RDF file ($file)!"
				unless -e $file;

			# FIXME: it might be wrong in some cases
			open(FF, $file) 
				or  croak "Couldn't open $file:\n$!";
			eval {
				$ret = $class->parse(*FF,$file_uri);
    			};
    			my $err = $@;
    			close(FF);
    			croak $err 	
				if $err;
		};
    		return $ret;
  	};
};

sub wget {
	my ($uri) = @_;

	croak "wget: input url is not an instance of URI"
                unless( (defined $uri) && ($uri->isa("URI")) );

        my $handle = &chat::open_port($uri->host,$uri->port);
        &chat::print($handle,"GET $uri HTTP/1.0\r\nUser-Agent: Alberto.Reggiori\@jrc.it\r\n\r\n");
        *S = *chat::S;
        my $line = <S>;
        if (!($line =~ m#^HTTP/(\d+)\.(\d+) (\d\d\d) (.+)$#)) {
        	&chat::close($handle);
                warn "Did not get HTTP/X.X header back...$line";
                return undef;
        };
        my $status = $3;
        my $reason = $4;
        if ( $status != 200 ) {
        	&chat::close($handle);
                warn "Error MSG returned from server: $status $reason\n";
                return undef;
        };
        while(<S>) {
        	chomp;
                last if m/^\s+$/;
        };
	return *S;
};

sub getAttributeValue {
	my ($expat,$attlist, $elName) = @_;

  	return undef
		if(!@{$attlist});

	my $n;
	for($n=0; $n<=$#{$attlist}; $n+=2) {
    		my $attname = $attlist->[$n];
    		return $attlist->[$n+1]
			if ($attname eq $elName);
  	};
  	return undef;
}

sub RDFXML_StartElementHandler {
	my $expat = shift;
	my $tag = shift;
	my @attlist = @_;

	my $xml_tag = $tag; # save it for later

	# Stack up the vNStodo namespace list - we could do it with map()
	my $sNamespace = $expat->namespace($tag);
	if(not(defined $sNamespace)) {			
		my ($prefix,$suffix) = split(':',$tag);
		if($prefix eq $XML::Parser::PenRDF::XMLSCHEMA) {
			$sNamespace = $XML::Parser::PenRDF::XMLSCHEMA;
			$tag = $sNamespace.$suffix; #xml:lang ---> xmllang
		} else {
			$expat->xpcroak("Unresolved namespace prefix '$prefix' for '$suffix'")
				if( (defined $prefix) && (defined $suffix) );
		};
	} else {
		$tag = $sNamespace.$tag;
	};
	push @{$expat->{PenRDF}->{vNStodo}},$sNamespace
		unless( (grep /$sNamespace/,@{$expat->{PenRDF}->{vNStodo}}) &&
			(grep /$sNamespace/,@{$expat->{PenRDF}->{vNSdone}}) &&
			($sNamespace ne $XML::Parser::PenRDF::RDF_SYNTAX_NS) &&
			($sNamespace ne $XML::Parser::PenRDF::RDF_SCHEMA_NS) );

	my $newElement;

	my $setScanModeElement = 0;
	if($expat->{PenRDF}->{scanMode} eq 'SKIPPING') {
		if( $tag eq $XML::Parser::PenRDF::RDFMS_RDF ) {
                        $expat->{PenRDF}->{scanMode} = 'RDF';
                        $setScanModeElement = 1;
                } elsif( $tag eq $XML::Parser::PenRDF::RDFMS_Description ) {
                        $expat->{PenRDF}->{scanMode} = 'TOP_DESCRIPTION';
                        $setScanModeElement = 1;
                };
	} elsif($expat->{PenRDF}->{scanMode} eq 'RDF') {
		$expat->{PenRDF}->{scanMode} = 'DESCRIPTION';
		$setScanModeElement = 1;
	};

	# speed up a bit...
	my $parseLiteral = (($expat->{PenRDF}->{scanMode} ne 'SKIPPING') && (parseLiteral($expat)));
	if( (!$parseLiteral) && ($expat->{PenRDF}->{scanMode} ne 'SKIPPING') ) {
		my $n;
		for($n=0; $n<=$#attlist; $n+=2) {
    			my $attname = $attlist[$n];
			my $namespace = $expat->namespace($attname);
			unless(defined $namespace) { #default namespace
				my ($prefix,$suffix) = split(':',$attname);
				if( (defined $prefix) && (defined $suffix) ) {
					if($prefix eq $XML::Parser::PenRDF::XMLSCHEMA) {
						$namespace = $XML::Parser::PenRDF::XMLSCHEMA;
						$attlist[$n] = $namespace.$suffix;
					} else {
						$expat->xpcroak("Unresolved namespace prefix '$prefix' for '$suffix'");
					};
				} else {
					if(	($attname eq 'resource') 	|| 
						($attname eq 'ID') 		|| 
						($attname eq 'about') 		|| 
						($attname eq 'aboutEach') 	|| 
						($attname eq 'aboutEachPrefix') ||
						($attname eq 'bagID')		||
						($attname eq 'parseType')	||
						($attname eq 'type') ) {
						#default to RDFMS
						$namespace = $XML::Parser::PenRDF::RDF_SYNTAX_NS;
					} else {
						$namespace = $sNamespace;
					};
					$attlist[$n]=$namespace.$attname;
				};
			} else {
				$attlist[$n] = $namespace.$attname;
			};
			push @{$expat->{PenRDF}->{vNStodo}}, $namespace
				unless( (grep /$namespace/,@{$expat->{PenRDF}->{vNStodo}}) &&
					(grep /$namespace/,@{$expat->{PenRDF}->{vNSdone}}) &&
					($namespace ne $XML::Parser::PenRDF::RDF_SYNTAX_NS) &&
					($namespace ne $XML::Parser::PenRDF::RDF_SCHEMA_NS) );
  		};
	};

	# If we have parseType="Literal" set earlier, this element
        # needs some additional attributes to make it stand-alone
        # piece of XML
	if($parseLiteral) {
		#ignored for the moment
		$newElement =  XML::Parser::PenRDF::Element->new($tag,\@attlist);
	} else {
		#....and probably Expat has already something like this.....
		$newElement =  XML::Parser::PenRDF::Element->new($tag,\@attlist);
	};

	$expat->{PenRDF}->{EXPECT_Element} = $newElement
		if($setScanModeElement);

	my $sLiteralValue;
	if($expat->{PenRDF}->{scanMode} ne 'SKIPPING') {

		# goes through the attributes of newElement to see
	 	# 1. if there are symbolic references to other nodes in the data model.
		# in which case they must be stored for later resolving with
		# resolveLater method (fix aboutEach aboutEachPrefix on streaming!!!)
		# 2. if there is an identity attribute, it is registered using
		# registerResource or registerID method. 
	
       		my $sResource = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_resource);
       	 	$newElement->{sResource} = makeAbsolute($expat,$sResource)
			if ($sResource);

        	my $sAboutEach = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_aboutEach);
        	$newElement->{sAboutEach} = $sAboutEach
			if ($sAboutEach =~ /^#/);

        	my $sAboutEachPrefix = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_aboutEachPrefix);
        	$newElement->{sAboutEachPrefix} = $sAboutEachPrefix
			if ($sAboutEachPrefix =~ /^#/);

        	my $sAbout = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_about);
        	if($sAbout) {
        		$newElement->{sAbout} = makeAbsolute($expat,$sAbout);
        	};

        	my $sBagID = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_bagID);

        	if ($sBagID) {
        		$newElement->{sBagID} = makeAbsolute($expat,$sBagID);
			$sBagID = $newElement->{sBagID};
        	};

        	my $sID = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_ID);
        	if ($sID) {
        		$newElement->{sID} = makeAbsolute($expat,$sID);
			$sID = $newElement->{sID};
        	};
		if( ($sAboutEach ne '') || ($sAboutEachPrefix ne '') ) {
			#any idea how to support it? caching and backrefs??
			$expat->xpcroak("aboutEach and aboutEachPrefix are not supported on stream parsing ");
		};

		if( ($sID ne '') && ($sAbout ne '') ) {
			$expat->xpcroak("A description block cannot use both 'ID' and 'about' attributes - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#idAboutAttr\">[6.5]</a>");
		};

		# Check parseType
		$sLiteralValue = getAttributeValue($expat,$newElement->{attlist},
				$XML::Parser::PenRDF::RDFMS_parseType);

		if ( (defined $sLiteralValue) && ($sLiteralValue ne 'Resource') ) {
			# This is the management of the element where
                	# parseType="Literal" appears
                	#
                	# You should notice RDF V1.0 conforming implementations
                	# must treat other values than Literal and Resource as
                	# Literal. This is why the condition is !equals("Resource")

			if(scalar(@{$expat->{PenRDF}->{elementStack}})>0) {
				my $e = $expat->{PenRDF}->{elementStack}->[$#{$expat->{PenRDF}->{elementStack}}];
				push @{$e->{children}},$newElement;
			};
        		# Place the new element into the stack
			push @{$expat->{PenRDF}->{elementStack}},$newElement;
			push @{$expat->{PenRDF}->{parseElementStack}},$newElement;
			$expat->{PenRDF}->{sLiteral} = '';

                	return;
		};

		if($parseLiteral) {
                	# This is the management of any element nested within
                	# a parseType="Literal" declaration

			#Trigger 'Start_XML_Literal' event
			my $start_literal = $expat->{Handlers}->{Start_XML_Literal}
				if(ref($expat->{Handlers}) =~ /HASH/);
			$expat->{PenRDF}->{sLiteral} .= &$start_literal($expat,$xml_tag,@attlist)
				if(defined $start_literal);
			push @{$expat->{PenRDF}->{elementStack}},$newElement;
			return;
        	};
        };

	# Update the containment hierarchy with the stack
	# Prevent hooking up of 1st level descriptions to the root element
	if (	(scalar(@{$expat->{PenRDF}->{elementStack}})>0) &&
		(!$setScanModeElement) ) {
		my $e = $expat->{PenRDF}->{elementStack}->[$#{$expat->{PenRDF}->{elementStack}}];
		push @{$e->{children}},$newElement;
	};

        # Place the new element into the stack
	push @{$expat->{PenRDF}->{elementStack}},$newElement;

	if ( (defined $sLiteralValue) && ($sLiteralValue eq 'Resource') ) {
		push @{$expat->{PenRDF}->{parseElementStack}},$newElement;
		$expat->{PenRDF}->{sLiteral} = '';

                # Since parseType="Resource" implies the following
                # production must match Description, let's create
                # an additional Description node here in the document tree.
                my $desc = XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description,
								\@attlist);

		if(scalar(@{$expat->{PenRDF}->{elementStack}})>0) {
                	my $e = $expat->{PenRDF}->{elementStack}->[$#{$expat->{PenRDF}->{elementStack}}];
                	push @{$e->{children}},$desc;
        	};
		push @{$expat->{PenRDF}->{elementStack}},$desc;
	};
};

sub RDFXML_EndElementHandler {
	my $expat = shift;
	my $tag = shift;

    	my $bParseLiteral = parseLiteral($expat);
    	$expat->{PenRDF}->{root} = pop @{$expat->{PenRDF}->{elementStack}};

	return
		if($expat->{PenRDF}->{scanMode} eq 'SKIPPING');

	if ($bParseLiteral) {
                my $pe = $expat->{PenRDF}->{parseElementStack}->[$#{$expat->{PenRDF}->{parseElementStack}}];
		if($pe != $expat->{PenRDF}->{root}) {
			#Trigger 'Stop_XML_Literal' event
			my $stop_literal = $expat->{Handlers}->{Stop_XML_Literal}
				if(ref($expat->{Handlers}) =~ /HASH/);
			$expat->{PenRDF}->{sLiteral} .= &$stop_literal($expat,$tag)
				if(defined $stop_literal);
		} else {
			# we would want resource because parseType="Literal" is text/xml (see RFC 2397)
			push @{$expat->{PenRDF}->{root}->{children}},XML::Parser::PenRDF::DataElement->new($expat->{PenRDF}->{sLiteral});
                	pop @{$expat->{PenRDF}->{parseElementStack}};
		};
	} elsif(parseResource($expat)) {
		# If we are doing parseType="Resource"
         	# we need to explore whether the next element in
         	# the stack is the closing element in which case
         	# we remove it as well (remember, there's an
         	# extra Description element to be removed)
		if(scalar(@{$expat->{PenRDF}->{elementStack}})>0) {
                	my $pe = $expat->{PenRDF}->{parseElementStack}->[$#{$expat->{PenRDF}->{parseElementStack}}];
                	my $e = $expat->{PenRDF}->{elementStack}->[$#{$expat->{PenRDF}->{elementStack}}];
            		if ($pe == $e) {
                		$e = pop @{$expat->{PenRDF}->{elementStack}};
                		pop @{$expat->{PenRDF}->{parseElementStack}};
            		};
		};
	};

	if($expat->{PenRDF}->{scanMode} eq 'RDF') {
		$expat->{PenRDF}->{scanMode} = 'SKIPPING';
		return;
	};

	# we are deep inside - I do not understand this by AR
	return
		if($expat->{PenRDF}->{EXPECT_Element} != $expat->{PenRDF}->{root});

	if($expat->{PenRDF}->{scanMode} eq 'TOP_DESCRIPTION') {
		processXML($expat,$expat->{PenRDF}->{EXPECT_Element});
		$expat->{PenRDF}->{scanMode} = 'SKIPPING';
	} elsif($expat->{PenRDF}->{scanMode} eq 'DESCRIPTION') {
		processXML($expat,$expat->{PenRDF}->{EXPECT_Element});
		$expat->{PenRDF}->{scanMode} = 'RDF';
	};
};

sub RDFXML_CharacterDataHandler {
	my $expat = shift;
	my $text = shift;

	if (parseLiteral($expat)) {
		#Trigger 'Char_Literal' event
		my $char_literal = $expat->{Handlers}->{Char_Literal}
			if(ref($expat->{Handlers}) =~ /HASH/);
		$expat->{PenRDF}->{sLiteral} .= &$char_literal($expat,$text)
			if(defined $char_literal);

        	return;
    	};

     	# Place all characters as Data instance to the containment
     	# hierarchy with the help of the stack.
    	my $e = $expat->{PenRDF}->{elementStack}->[$#{$expat->{PenRDF}->{elementStack}}];

	# Determine whether the previous event was for
	# characters. If so, update the Data node contents.
	# A&amp;B would otherwise result in three
	# separate Data nodes in the parse tree
	my $bHasData = 0;
	my $dN;
	my $dataNode;
	foreach $dN (@{$e->{children}}) {
		if($dN->isa('XML::Parser::PenRDF::DataElement')) {
			$bHasData = 1;
			$dataNode=$dN;
			last;
		};
      	};

     	# Warning: this is not correct procedure according to XML spec.
	# All whitespace matters!

	# trim text
	my $trimtext = $text;
	$trimtext =~ s/^([ ])+//g;
	$trimtext =~ s/([ ])+$//g;

	# FIXME? it seems not really correct $preserveWhiteSpace
	my $preserveWhiteSpace=0;
	my $size = scalar(@{$expat->{PenRDF}->{elementStack}});
	if ($size>0) {	
		my $e = $expat->{PenRDF}->{elementStack}->[$size-1];
    		my $sParseType = getAttributeValue($expat, $e->{attlist},$XML::Parser::PenRDF::XML_space);
		$preserveWhiteSpace=($XML::Parser::PenRDF::XML_space_preserve eq $sParseType)
			if(defined $sParseType);
	};
	if( (length($trimtext)>0) || ($preserveWhiteSpace) ) {
		if(!$bHasData) {
			return
				if(	(!$preserveWhiteSpace) 	&& 
					($text =~ /^\s+/) 	&& 
					($text =~ /\s+$/) );
        		push @{$e->{children}},XML::Parser::PenRDF::DataElement->new($text);
		} else {
        		$dataNode->{sContent} .= $text;
			#not nice to see it here.....I know ;-)
			$dataNode->{tag} = "[DATA: " . $dataNode->{sContent} . "]";
		};
	};
};

sub processXML {
	my ($expat,$ele) = @_;

	if($ele->{tag} eq $XML::Parser::PenRDF::RDFMS_RDF) {
		my $c;
		foreach $c (@{$ele->{children}}) {
			if($c->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
				processDescription($expat,$c,0,
					$expat->{PenRDF}->{bCreateBags}, $expat->{PenRDF}->{bCreateBags});
			} elsif( 	($c->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
					($c->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
					($c->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag)	)  {
				processContainer($expat,$c);
			#strange checking here....
			} elsif( 	(!($c->{tag} eq $XML::Parser::PenRDF::RDFMS_resource)) && 
					(length($c->{tag})>0) ) {
				processTypedNode($expat,$c);
			};
		};
	} elsif($ele->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
		processDescription($expat,$ele,0,
				$expat->{PenRDF}->{bCreateBags}, $expat->{PenRDF}->{bCreateBags});
	} else {
		processTypedNode($expat,$ele);
	};

	# Recursively call myself to go through all the schemas
	#....even more similar to the PEN one ;-)
	if ($expat->{PenRDF}->{bFetchSchemas}) {
		my $sURI;
		while($sURI=shift @{$expat->{PenRDF}->{vNStodo}}) {
			push @{$expat->{PenRDF}->{vNSdone}},$sURI;

			# Merge all the schemas :-)
			#
			# There is a problem that the same XML::Parser::Expat instace can not
			# parse different XML chuncks :-( 
			# ...but we need to merge the triples somehow....
			#
			# Run Expat
        		my $third = new XML::Parser::Expat(@{$expat->{parser_parameters}});

			# a bit tricky here....but should work ;-)
        		$third->{PenRDF} = $expat->{PenRDF};
			$third->{PenRDF}->{sSource}=setSource($sURI);

        		$third->setHandlers(    Start => \&RDFXML_StartElementHandler,
                        	        	End => \&RDFXML_EndElementHandler,
                               			Char => \&RDFXML_CharacterDataHandler );

			my $file_uri;
			if(-e $sURI) {
				$file_uri= URI::file->new_abs($sURI);
			} elsif(!($sURI =~ /^[\w]+:/)) {
				$file_uri= URI::file->new_abs($sURI,URI::file->cwd);
			} else {
				$file_uri= URI->new($sURI);
			};
			if (	(defined $file_uri->scheme)	&&
				($file_uri->scheme ne 'file') ) {
				my $uri = new URI($sURI);
				my $wget_handle = wget($uri);
                        	if(defined $wget_handle) {
                                	eval {
                                        	my $ret = $third->parse($wget_handle, $uri);
                                	};
                                	my $err = $@;
                                	croak $err
                                        	if $err;
                        	} else {
					croak "Cannot fetch '$uri'";
				};
			} else {
				my $file= $file_uri->file;

				croak "Unable to open input RDF file ($file)!"
					unless -e $file;
				open(FF, $file)
					or  croak "Couldn't open $file:\n$!";
				eval {
					my $ret = $third->parse(*FF, $file_uri);
    				};
    				my $err = $@;
    				close(FF);
    				croak $err 	
					if $err;
			};

			# then trigger the nice thing ;-)
			# in the $first pass streaming the first trigger is on "Description"
			processXML($third->{PenRDF}->{root});
      		};
    	};
};

sub processDescription {
	my ($expat,$ele,$inPredicate,$reify,$createBag) = @_;

	# Return immediately if the description has already been managed
	return $ele->{sID}
		if($ele->{bDone});

	my $iChildCount=1;
	my $bOnce=1;
	
	# Determine first all relevant values
	my ($sID,$sBagid,$sAbout,$sAboutEach,$sAboutEachPrefix) = (
									$ele->{sID},
									$ele->{sBagID},
									$ele->{sAbout},
									$ele->{sAboutEach},
									$ele->{sAboutEachPrefix}
									);
	my $target = (defined $ele->{vTargets}->[0]) ? $ele->{vTargets}->[0] : undef;

	my $targetIsContainer=0;
	my $sTargetAbout='';
	my $sTargetBagid='';
	my $sTargetID='';

	# Determine what the target of the Description reference is
	if (defined $target) {
      		my $sTargetAbout = $target->{sAbout};
      		my $sTargetID    = $target->{sID};
      		my $sTargetBagid = $target->{sBagID};

       		# Target is collection if
       		# 1. it is identified with bagID attribute
       		# 2. it is identified with ID attribute and is a collection
      		if ( ((defined $sTargetBagid) && ($sTargetBagid ne '')) && 
			((defined $sAbout) && ($sAbout ne '')) ) {
			# skip '#' sign??
        		$targetIsContainer = ($sAbout =~ /^.$sTargetBagid/);
      		} else {
        		if (	((defined $sTargetID) && ($sTargetID ne '')) &&
            			((defined $sAbout) && ($sAbout ne '')) &&
				($sAbout =~ /^.$sTargetID/) &&
				( 	($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
                                        ($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
                                        ($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) )	)  {
          			$targetIsContainer = 1;
        		};
      		};
    	};

	# Check if there are properties encoded using the abbreviated syntax
	expandAttributes($expat,$ele,$ele,0);

	# Manage the aboutEach attribute here
	if( ((defined $sAboutEach) && ($sAboutEach ne '')) && (defined $target) ) {
      		if( 	($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
                        ($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
                        ($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) ) {
			my $ele1;
			foreach $ele1 (@{$target->{children}}) {
          			if( 	($ele1->{tag} =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/) &&
					( ($ele1->{tag} =~ /li$/) || ($ele1->{tag} =~ /_/) ) ) {
            				my $sResource = $ele1->{sResource};
             				# Manage <li resource="..." /> case
            				if((defined $sResource) && ($sResource ne '')) {
              					my $newDescription =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
						$newDescription->{sAbout} = $sResource;
					
						my $ele2;
						foreach $ele2 (@{$ele->{children}}) {
							if (defined $newDescription) {
                  						push @{$newDescription->{children}},$ele2;
                					};
              					};
                				processDescription($expat,$newDescription,0,0,0)
							if (defined $newDescription);
            				} else {
               					# Otherwise we have a structured value inside <li>
              					# loop through the children of <li>
              					# (can be only one)
						my $ele2;
						foreach $ele2 (@{$ele1->{children}}) {
                					# loop through the items in the
                					# description with aboutEach
                					# and add them to the target
              						my $newNode =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
							my $ele3;
							foreach $ele3 (@{$ele->{children}}) {
								if (defined $newNode) {
                  							push @{$newNode->{children}},$ele3;
								};
              						};
                					push @{$newNode->{vTargets}},$ele2;

               						processDescription($expat,$newNode,1,0,0);
       						};
       					};
				} elsif( 	(!($ele1->{tag} eq $XML::Parser::PenRDF::RDFMS_resource)) &&
						(length($ele1->{tag})>0) ) {
              				my $newNode =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
					my $ele2;
					foreach $ele2 (@{$ele->{children}}) {
						if (defined $newNode) {
                  					push @{$newNode->{children}},$ele2;
						};
              				};
                			push @{$newNode->{vTargets}},$ele1;
                			processDescription($expat,$newNode,1,0,0);
          			};
        		};
		} elsif($target->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
                	processDescription($expat,$target,0,$reify,$createBag);
			my $ele1;
			foreach $ele1 (@{$target->{children}}) {
              			my $newNode =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
				my $ele2;
				foreach $ele2 (@{$ele->{children}}) {
					if (defined $newNode) {
                  				push @{$newNode->{children}},$ele2;
					};
              			};
                		push @{$newNode->{vTargets}},$ele1;
                		processDescription($expat,$newNode,1,0,0);
        		};
	 	};
      		return undef;
    	};

	# Manage the aboutEachPrefix attribute
	if((defined $sAboutEachPrefix) && ($sAboutEachPrefix ne '')) {
      		if (defined $target) {
			foreach $target (@{$ele->{vTargets}}) {
          			$sTargetAbout = $target->{sAbout};
              			my $newDescription =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
				$newDescription->{ sAbout } = $sTargetAbout;
				my $ele2;
				foreach $ele2 (@{$ele->{children}}) {
					if (defined $newDescription) {
                  				push @{$newDescription->{children}},$ele2;
					};
          			};
                		processDescription($expat,$newDescription,0,0,0);
        		};
      		};
      		return undef;
    	};

	# Enumerate through the children
	my $paCounter = 1;
	my $n;
	foreach $n (@{$ele->{children}}) {
		if($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
        		$expat->xpcroak("Cannot nest a Description inside another Description");
      		} elsif( 	($n->{tag} =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/) &&
					( ($n->{tag} =~ /li$/) || ($n->{tag} =~ /_/) ) ) {
        		processListItem($expat,$sID,$n,$paCounter);
        		$paCounter++;
      		} elsif( 	($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
                                ($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
                                ($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) ) {
        		$expat->xpcroak("Cannot nest a container (Bag/Alt/Seq) inside a Description");
      		} elsif( 	(!($n->{tag} eq $XML::Parser::PenRDF::RDFMS_resource)) && 
				(length($n->{tag})>0) ) {
        		my $sChildID;
			if ( (defined $target) && ($targetIsContainer) ) {
          			$sChildID = processPredicate($expat,$n,$ele,
                                       ((defined $target->{sBagID}) ? $target->{sBagID} : $target->{sID}),0);
          			$ele->{sID} = makeAbsolute($expat,$sChildID);
				$createBag=0;
        		} elsif(defined $target) {
          			$sChildID = processPredicate($expat,$n,$ele,
                                       ((defined $target->{sBagID}) ? $target->{sBagID} : $target->{sID}),$reify);
          			$ele->{sID} = makeAbsolute($expat,$sChildID);
        		} elsif( (not(defined $target)) && (!($inPredicate)) ) {
          			$ele->{sID} = newReificationID($expat)
          				if(not(((defined $ele->{sID}) && ($ele->{sID} ne ''))));
          			if (not(((defined $sAbout) && ($sAbout ne '')))) {
            				if ((defined $sID) && ($sID ne '')) {
						$sAbout=$sID;
					} else {
						$sAbout=$ele->{sID};
					};
				};
          			$sChildID = processPredicate($expat,$n,$ele,
						$sAbout,
						( ((defined $sBagid) && ($sBagid ne '')) ? 1 : $reify));
        		} elsif( (not(defined $target)) && ($inPredicate) ) {
          			if (not(((defined $sAbout) && ($sAbout ne '')))) {
            				if ((defined $sID) && ($sID ne '')) {
          					$ele->{sID} = makeAbsolute($expat,$sID);
						$sAbout=$sID;
					} else {
          					$ele->{sID} = newReificationID($expat)
          						if(not(((defined $ele->{sID}) && ($ele->{sID} ne ''))));
						$sAbout=$ele->{sID};
					};
				} else {
          				$ele->{sID} = $sAbout;
				};
					
          			$sChildID = processPredicate($expat,$n,$ele,$sAbout,0);
        		};

                        # Each Description block creates also a Bag node which
                        # has links to all properties within the block IF
                        # the bCreateBags variable is true
        		if( ((defined $sBagid) && ($sBagid ne '')) || ($expat->{PenRDF}->{bCreateBags} && $createBag) ) {
          			my $sNamespace = $XML::Parser::PenRDF::RDF_SYNTAX_NS;
          			# do only once and only if there is a child
          			if( ($bOnce) && ((defined $sChildID) && ($sChildID ne '')) ) {
            				$bOnce = 0;
              				$ele->{sBagID} = newReificationID($expat)
            					if(not(((defined $ele->{sBagID}) && ($ele->{sBagID} ne ''))));
          				$ele->{sID} = makeAbsolute($expat,$ele->{sBagID})
          					if(not(((defined $ele->{sID}) && ($ele->{sID} ne ''))));

			            	addTriple(	$expat,
							$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
                       					$expat->{PenRDF}->{nodeFactory}->createResource($ele->{sBagID}),
							$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_Bag)
						);
          			};
				if ((defined $sChildID) && ($sChildID ne '')) {
			            	addTriple(	$expat,
							$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iChildCount),
                       					$expat->{PenRDF}->{nodeFactory}->createResource($ele->{sBagID}),
							$expat->{PenRDF}->{nodeFactory}->createResource($sChildID)
						 );
            				$iChildCount++;
          			};
        		};
      		};
    	};
	$ele->{bDone} = 1;
	
	return $ele->{sID};
};

# we could use URI and XPath modules to validate and normalise the subject, predicate, object
# Use XPath/XPointer for literals could be cool to have one unique uri thing
sub addTriple {
	my ($expat,$predicate,$subject,$object) = @_;

#print STDERR "addTriple('$predicate','$subject','$object')",((caller)[2]),"\n";

	carp "Predicate null when subject=".$subject." and object=".$object
		unless(defined $predicate);

	carp "Object null when predicate=".$predicate." and subject=".$subject
		unless(defined $object);
	$subject = $expat->{PenRDF}->{nodeFactory}->createResource($expat->{PenRDF}->{sSource})
		unless( (defined $subject) && (length($subject)>0) );

	# If there is no subject (about=""), then use the URI/filename where
	# the RDF description came from 

	#Trigger 'Assert' event
        my $assert = $expat->{Handlers}->{Assert}
		if(ref($expat->{Handlers}) =~ /HASH/);
        if (defined($assert)) {
        	return &$assert($expat, 
				$expat->{PenRDF}->{nodeFactory}->createStatement($subject,$predicate,$object) );
	} else {
		return undef;
	};
};

sub newReificationID {
	my ($expat) = @_;

	$expat->{PenRDF}->{iReificationCounter}++;

	my $reifStr;
	if( (defined $expat->{PenRDF}->{sSource}) && ($expat->{PenRDF}->{sSource} ne '') ) {
		$reifStr=makeAbsolute($expat,"genid".$expat->{PenRDF}->{iReificationCounter});
	} else {
		$reifStr="genid".$expat->{PenRDF}->{iReificationCounter};
	};
	return $reifStr;
};

sub processTypedNode {
	my ($expat,$typedNode) = @_;

	my $sID = $typedNode->{sID};
	my $sBagID = $typedNode->{sBagID};
	my $sAbout = $typedNode->{sAbout};

	my $target = (defined $typedNode->{vTargets}->[0]) ? $typedNode->{vTargets}->[0] : undef;

    	my $sAboutEach = $typedNode->{sAboutEach};
	my $sAboutEachPrefix = $typedNode->{sAboutEachPrefix};

	if ( (defined $typedNode->{sResource}) && ($typedNode->{sResource} ne '') ) {
      		$expat->xpcroak("'resource' attribute not allowed for a typedNode '".$typedNode->{tag}."' - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#typedNode\">[6.13]</a>");
	};

	# We are going to manage this typedNode using the processDescription
	# routine later on. Before that, place all properties encoded as
	# attributes to separate child nodes.
	my $n;
	for($n=0; $n<=$#{$typedNode->{attlist}}; $n+=2) {
    		my $sAttribute = $typedNode->{attlist}->[$n];
    		my $sValue = getAttributeValue($expat, $typedNode->{attlist},$sAttribute);
		$sValue =~ s/^([ ])+//g;
		$sValue =~ s/([ ])+$//g;

		if ( 	(!($sAttribute =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/)) &&
			(!($sAttribute =~ /^$XML::Parser::PenRDF::XMLSCHEMA/)) ) {
        		if(length($sValue) > 0) {
              			my $newPredicate =  XML::Parser::PenRDF::Element->new(
							$sAttribute,[
						$XML::Parser::PenRDF::RDFMS_ID, 
						(((defined $sAbout) && ($sAbout ne '')) ? $sAbout : $sID),
						$XML::Parser::PenRDF::RDFMS_bagID, 
						$sBagID
								]);
				
				my $newData =  XML::Parser::PenRDF::DataElement->new($sValue);
				push @{$newPredicate->{children}},$newData;
				push @{$typedNode->{children}},$newPredicate;

				# removeAttribute
				my @rr;
				my $i;
				for($i=0; $i<=$#{$typedNode->{attlist}}; $i++) {
					if($typedNode->{attlist}->[$i] eq $sAttribute) {
						$i++;
					} else {
						push @rr,$typedNode->{attlist}->[$i];
					};
				};
				$typedNode->{attlist} = \@rr;
        		};
      		};
    	};

	my $sObject;
    	if(defined $target) {
		$sObject = ( (((defined $target->{sBagID}) && ($target->{sBagID} ne ''))) ? $target->{sBagID} : $target->{sID});
	} elsif((defined $sAbout) && ($sAbout ne '')) {
      		$sObject = $sAbout;
    	} elsif((defined $sID) && ($sID ne '')) {
      		$sObject = $sID;
    	} else {
      		$sObject = newReificationID($expat);
	};

	$typedNode->{sID} = makeAbsolute($expat,$sObject);

	# special case: should the typedNode have aboutEach attribute,
	# the type predicate should distribute to pointed
	# collection also -> create a child node to the typedNode
	if ( 	((defined $sAboutEach) && ($sAboutEach ne '')) &&
        	(scalar(@{$typedNode->{vTargets}})>0) ) {
              		my $newPredicate =  XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_type);
			my $newData = XML::Parser::PenRDF::DataElement->new($typedNode->{tag});
			push @{$newPredicate->{children}},$newData;
			push @{$typedNode->{children}},$newPredicate;
    	} else {
      		addTriple(	$expat,
				$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
				$expat->{PenRDF}->{nodeFactory}->createResource($typedNode->{sID}),
				$expat->{PenRDF}->{nodeFactory}->createResource($typedNode->{tag})
			);
    	};

    	my $sDesc = processDescription($expat,$typedNode, 0, 0, 0);

    	return $sObject;
};

sub processContainer {
	my ($expat,$n) = @_;
	
	my $sID = $n->{sID};
      	$sID = $n->{sAbout}
    		unless((defined $sID) && ($sID ne ''));
      	$sID = newReificationID($expat)
    		unless((defined $sID) && ($sID ne ''));

     	# Do the instantiation only once
	if(!($n->{bDone})) {
      		if($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) {
			addTriple(	$expat,
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
                       			$expat->{PenRDF}->{nodeFactory}->createResource($sID),
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_Seq) 
				);
      		} elsif($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) {
			addTriple(	$expat,
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
                       			$expat->{PenRDF}->{nodeFactory}->createResource($sID),
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_Alt) 
				);
      		} elsif($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) {
			addTriple(	$expat,
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
                       			$expat->{PenRDF}->{nodeFactory}->createResource($sID),
					$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_Bag) 
				);
      		};
		$n->{bDone} = 1;
    	};

	expandAttributes($expat,$n,$n,0);

	if( 	(scalar(@{$n->{children}})<=0) &&
      		($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ) {
      		$expat->xpcroak("An RDF:Alt container must have at least one nested listitem");
    	};

	my $iCounter = 1;
	my $n2;
	foreach $n2 (@{$n->{children}}) {
		if (	($n2->{tag} =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/) &&
			( ($n2->{tag} =~ /li$/) || ($n2->{tag} =~ /_/) ) ) {
			processListItem($expat,$sID, $n2, $iCounter);
        		$iCounter++;
      		} else {
        		$expat->xpcroak("Cannot nest ".$n2->{tag}." inside a container (Bag/Alt/Seq)");
      		};
    	};
	return $sID;
};

sub processListItem {
	my ($expat,$sID,$listitem,$iCounter) = @_;

	# Two different cases for
     	# 1. LI element without content (resource available)
     	# 2. LI element with content (resource unavailable)

	my $sResource = $listitem->{sResource};
    	if((defined $sResource) && ($sResource ne '')) {
		addTriple(	$expat,
				$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iCounter),
                       		$expat->{PenRDF}->{nodeFactory}->createResource($sID),
				$expat->{PenRDF}->{nodeFactory}->createResource($sResource)
			 );
      		# validity checking
      		if (scalar(@{$listitem->{children}})>0) {
        		$expat->xpcroak("Listitem with 'resource' attribute cannot have child nodes - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#referencedItem\">[6.29]</a>");
      		};
      		$listitem->{sID} = makeAbsolute($expat,$sResource);
    	} else {
		my $n;
		foreach $n (@{$listitem->{children}}) {
        		if($n->isa('XML::Parser::PenRDF::DataElement')) { #isData?
				addTriple(	$expat,
						$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iCounter),
                       				$expat->{PenRDF}->{nodeFactory}->createResource($sID),
						$expat->{PenRDF}->{nodeFactory}->createLiteral($n->{sContent})
					 ); #here is Literal - how to handle it???
        		} elsif($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
          			my $sNodeID = processDescription($expat,$n, 0,1, 0);
				addTriple(	$expat,
						$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iCounter),
                       				$expat->{PenRDF}->{nodeFactory}->createResource($sID),
						$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID)
					) ;

				$listitem->{sID} = makeAbsolute($expat,$sNodeID);
        		} elsif (     ($n->{tag} =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/) &&
                                        ( ($n->{tag} =~ /li$/) || ($n->{tag} =~ /_/) ) ) {
        			$expat->xpcroak("Cannot nest a listitem inside another listitem");
        		} elsif( 	($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
                                	($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
                                	($n->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) ) {
				processContainer($n);
				addTriple(	$expat,
						$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iCounter),
                       				$expat->{PenRDF}->{nodeFactory}->createResource($sID),
						$expat->{PenRDF}->{nodeFactory}->createResource($n->{sID})
					);
      			} elsif( 	(!($n->{tag} eq $XML::Parser::PenRDF::RDFMS_resource)) && 
					(length($n->{tag})>0) ) {
				my $sNodeID = processTypedNode($expat,$n);
				addTriple(	$expat,
						$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."_".$iCounter),
                       				$expat->{PenRDF}->{nodeFactory}->createResource($sID),
						$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID)
					);
        		};
      		};
	};
};

# processPredicate handles all elements not defined as special
# RDF elements. <tt>predicate</tt> has either <tt>resource()</tt> or a single child
sub processPredicate {
	my ($expat,$predicate,$description,$sTarget,$reify) = @_;

#print "processPredicate($predicate,$description,$sTarget,$reify)",((caller)[2]),"\n";

	my $sStatementID = $predicate->{sID};
	my $sBagID       = $predicate->{sBagID};
    	my $sResource    = $predicate->{sResource};

     	# If a predicate has other attributes than rdf:ID, rdf:bagID,
     	# or xmlns... -> generate new triples according to the spec.
     	# (See end of Section 6)

	# this new element may not be needed
        my $d = XML::Parser::PenRDF::Element->new($XML::Parser::PenRDF::RDFMS_Description);
    	if (expandAttributes($expat,$d,$predicate,1)) {
      		# error checking
      		if(scalar(@{$predicate->{children}})>0) {
        		$expat->xpcroak($predicate->{tag}." must be an empty element since it uses propAttr grammar production - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#propertyElt\">[6.12]</a>");
        		#cluck($predicate->{tag}." must be an empty element since it uses propAttr grammar production - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#propertyElt\">[6.12]</a>");
        		return undef;
      		};

      		# determine the 'about' part for the new statements
      		if ((defined $sStatementID) && ($sStatementID ne '')) {
        		push @{$d->{attlist}},$XML::Parser::PenRDF::RDFMS_about;
        		push @{$d->{attlist}},$sStatementID;

                        # make rdf:ID the value of the predicate
			push @{$predicate->{children}}, XML::Parser::PenRDF::DataElement->new($sStatementID);
      		} elsif ((defined $sResource) && ($sResource ne '')) {
        		push @{$d->{attlist}},$XML::Parser::PenRDF::RDFMS_about;
        		push @{$d->{attlist}},$sResource;
      		} else {
			$sStatementID = newReificationID($expat);
        		push @{$d->{attlist}},$XML::Parser::PenRDF::RDFMS_about;
        		push @{$d->{attlist}},$sStatementID;
      		};

		if ((defined $sBagID) && ($sBagID ne '')) {
        		push @{$d->{attlist}},$XML::Parser::PenRDF::RDFMS_bagID;
        		push @{$d->{attlist}},$sBagID;
        		$d->{sBagID} = $sBagID;
      		};

          	processDescription($expat,$d, 0,0,$expat->{PenRDF}->{bCreateBags});
    	};

	# Tricky part: if the resource attribute is present for a predicate
	# AND there are no children, the value of the predicate is either
	# 1. the URI in the resource attribute OR
	# 2. the node ID of the resolved #resource attribute
	my $predicate_target = (defined $predicate->{vTargets}->[0]) ? $predicate->{vTargets}->[0] : undef;
    	if( ((defined $sResource) && ($sResource ne '')) && (scalar(@{$predicate->{children}})<=0) ) {
      		if (not(defined $predicate_target)) {
        		if ($reify) {
          			$sStatementID = reify(	$expat,
							$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
							$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
							$expat->{PenRDF}->{nodeFactory}->createResource($sResource),
							$predicate->{sID});
				$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        		} else {
          			addOrder( 	$expat,
						$predicate, 0, undef,
                     				$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($sResource)
					);
        		};
      		} else {
        		if ($reify) {
          			$sStatementID = reify(	$expat, 
							$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
							$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
							$expat->{PenRDF}->{nodeFactory}->createResource($predicate_target->{sID}),
							$predicate->{sID});
				$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        		} else {
          			addOrder( 	$expat,
						$predicate, 0, undef,
                     				$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($predicate_target->{sID})
					);
        		};
      		};
      		return $predicate->{sID};
    	};
                                    
	# Does this predicate make a reference somewhere using the <i>sResource</i> attribute
    	if ( ((defined $sResource) && ($sResource ne '')) && (defined $predicate_target) ) {
      		$sStatementID = processDescription ($expat,$predicate_target,1,0,0);

        	if ($reify) {
          		$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
						$expat->{PenRDF}->{nodeFactory}->createResource($sStatementID),
						$predicate->{sID});
			$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        	} else {
          		addOrder( 	$expat,
					$predicate, 0, undef,
                     			$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sStatementID)
				 );
        	};
		return $sStatementID;
    	};

	# Before looping through the children, let's check
	# if there are any. If not, the value of the predicate is an anonymous node
    	my $sObject = newReificationID($expat);
	if (scalar(@{$predicate->{children}})<=0) {
        	if ($reify) {
          		$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
						$expat->{PenRDF}->{nodeFactory}->createResource($sObject),
						$predicate->{sID});
        	} else {
          		addOrder( 	$expat,
					$predicate, 0, undef,
                     			$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sObject)
				 );
		};
	};
	my $previousStatement;
	my $order = 0;
	my $n2;
	foreach $n2 (@{$predicate->{children}}) {
      		# FIXME: we are not requiring this for the sake of experiment
		# $expat->xpcroak("Only one node allowed inside a predicate (Extra node is ". $n2->{tag} .") - see <a href=\"http://www.w3.org/TR/REC-rdf-syntax/#propertyElt\">[6.12]</a>");

		if ($n2->{tag} eq $XML::Parser::PenRDF::RDFMS_Description) {
        		my $d2 = $n2;
        		$sStatementID = processDescription ($expat,$d2, 1,0,0);
        		$d2->{sID} = makeAbsolute($expat,$sStatementID);
        		if ($reify) {
          			$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
						$expat->{PenRDF}->{nodeFactory}->createResource($sStatementID),
						$predicate->{sID});
        		} else {
          			addOrder( 	$expat,
						$predicate, ++$order, $previousStatement,
                     				$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     				$expat->{PenRDF}->{nodeFactory}->createResource($sStatementID)
					 );
        		};
      		} elsif ($n2->isa('XML::Parser::PenRDF::DataElement')) {
			# We've got real data
        		my $sValue = $n2->{sContent};

                        # Only if the content is not empty PCDATA (whitespace that is), print the triple
        		if ($reify) {
          			$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
						$expat->{PenRDF}->{nodeFactory}->createLiteral($sValue),
						$predicate->{sID}); #ignore isXML
				$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        		} else {
          			$previousStatement = addOrder ( 	$expat,
									$predicate, ++$order, 
									$previousStatement,
                     							$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     							$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
									$expat->{PenRDF}->{nodeFactory}->createLiteral($sValue)
								); #ignore isXML
        		};
      		} elsif( 	($n2->{tag} eq $XML::Parser::PenRDF::RDFMS_Seq) ||
                            	($n2->{tag} eq $XML::Parser::PenRDF::RDFMS_Alt) ||
                                ($n2->{tag} eq $XML::Parser::PenRDF::RDFMS_Bag) ) {

			my $sCollectionID = processContainer($expat,$n2);
        		$sStatementID = $sCollectionID;

			# Attach the collection to the current predicate
			my $description_target = (defined $description->{vTargets}->[0]) ? $description->{vTargets}->[0] : undef;
        		if (defined $description_target) {
        			if ($reify) {
          				$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($description_target->{sAbout}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sCollectionID),
						$predicate->{sID});
					$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        			} else {
					addTriple(	$expat,
							$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
							$expat->{PenRDF}->{nodeFactory}->createResource($description_target->{sAbout}),
							$expat->{PenRDF}->{nodeFactory}->createResource($sCollectionID)
						 );
        			};
        		} else {
        			if ($reify) {
          				$sStatementID = reify(	$expat, 
						$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
						$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
						$expat->{PenRDF}->{nodeFactory}->createResource($sCollectionID),
						$predicate->{sID});
					$predicate->{sID} = makeAbsolute($expat,$sStatementID);
        			} else {
					addTriple(	$expat,
							$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
							$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
							$expat->{PenRDF}->{nodeFactory}->createResource($sCollectionID)
						 );
        			};
        		};
      		} elsif( 	(!($n2->{tag} eq $XML::Parser::PenRDF::RDFMS_resource)) && 
				(length($n2->{tag})>0) ) {
        		$sStatementID = processTypedNode($expat,$n2);
          		$previousStatement = addOrder ( 	$expat,
					$predicate, ++$order, $previousStatement,
                     			$expat->{PenRDF}->{nodeFactory}->createResource($predicate->{tag}),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sTarget),
                     			$expat->{PenRDF}->{nodeFactory}->createResource($sStatementID)
					 );
      		};
    	};
	return $sStatementID;
};

sub reify {
	my ($expat,$predicate,$subject,$object,$sNodeID) = @_;

	$sNodeID = newReificationID($expat)
    		if(not(((defined $sNodeID) && ($sNodeID ne ''))));

     	# The original statement must remain in the data model
    	addTriple($expat,$predicate, $subject, $object);

	# Do not reify reifyd properties
    	if (	($predicate eq $XML::Parser::PenRDF::RDFMS_subject) ||
    		($predicate eq $XML::Parser::PenRDF::RDFMS_predicate) ||
    		($predicate eq $XML::Parser::PenRDF::RDFMS_object) ||
    		($predicate eq $XML::Parser::PenRDF::RDFMS_type) ) {
      		return undef;
    	};

	# Reify by creating 4 new triples
    	addTriple(	$expat,
			$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_predicate),
			$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID),
			$predicate);

	$subject = (length($subject) == 0) ? $expat->{PenRDF}->{sSource} : $subject;
    	addTriple(	$expat,
			$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_subject),
			$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID),	
			$expat->{PenRDF}->{nodeFactory}->createLiteral($subject)
			);

    	addTriple(	$expat,
			$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_object),
			$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID),
			$object);

    	addTriple(	$expat,
			$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_type),
			$expat->{PenRDF}->{nodeFactory}->createResource($sNodeID),
			$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDFMS_Statement)
			);

	return $sNodeID;
};

# check for rdf:order and rdf:backwardOrder
# orderInt == 1 is ignored, it is added on the second call with previousStatement (orderInt == 2)
# orderInt < 1 is completely ignored 
sub addOrder {
	my ($expat,$predicate,$orderInt,$previousStatement,$p,$s,$o) = @_;

#print STDERR "addOrder('",$p->getLabel,"','",$s->getLabel,"','",$o->getLabel,"')",((caller)[2]),"\n";

	if($orderInt == 2) {
      		# add order of the previous statement if any
		addTriple(	$expat,
				$previousStatement,
				$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."order"),
				$expat->{PenRDF}->{nodeFactory}->createLiteral("1")
			);
    	};

	# "createStatement" must be handled within the callback :-)
    	my $st = $expat->{PenRDF}->{nodeFactory}->createStatement($s,$p,$o);
	addTriple($expat,$p,$s,$o);

	my $order = getAttributeValue(	$expat, $predicate->{attlist},
					$XML::Parser::PenRDF::RDF_SYNTAX_NS."order");
	my $backwardOrder =getAttributeValue(	$expat, $predicate->{attlist},
					$XML::Parser::PenRDF::RDF_SYNTAX_NS."backwardOrder");

    	if( (defined $previousStatement) && (defined $order) ) {
		# two statements cannot have the same forward order
      		$expat->xpcroak("two statements " . $previousStatement ." and " . $st . " cannot have the same forward order");
    	};

      	$order = $orderInt
    		if( (not(defined $order)) && ($orderInt > 1) );
    	if(defined $order) {
		addTriple(	$expat,
				$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."order"),
				$st,
				$expat->{PenRDF}->{nodeFactory}->createLiteral($order)
			);
    	};
    	if(defined $backwardOrder) {
		addTriple(	$expat,
				$expat->{PenRDF}->{nodeFactory}->createResource($XML::Parser::PenRDF::RDF_SYNTAX_NS."backwardOrder"),
				$st,
				$expat->{PenRDF}->{nodeFactory}->createLiteral($backwardOrder)
			);
    	};

	return $st;
};

# Take an element <i>ele</i> with its parent element <i>parent</i>
# and evaluate all its attributes to see if they are non-RDF specific
# and non-XML specific in which case they must become children of
# the <i>ele</i> node.
sub expandAttributes {
	my ($expat,$parent,$ele,$predicateNode) = @_;

#print "expandAttributes($parent,$ele->{tag},$predicateNode)",((caller)[2]),"\n";

	my $foundAbbreviation = 0;
	
  	my $count=0;
  	while ($count<=$#{$ele->{attlist}}) {
    		my $sAttribute = $ele->{attlist}->[$count++];
    		my $sValue = getAttributeValue($expat, $ele->{attlist},$sAttribute);
		$count++;
      		if ($sAttribute =~ /^$XML::Parser::PenRDF::XMLSCHEMA/) {
        		# expands after parsing, that's why it is useless here... :(
           		# because of concatenation without : inbetween
			# ...there was something more here to do....
        		next;
      		};

      		# exception: expand rdf:value
		my $notit=$XML::Parser::PenRDF::RDF_SYNTAX_NS .'_';
      		if (	($sAttribute =~ /^$XML::Parser::PenRDF::RDF_SYNTAX_NS/) &&
          		(!($sAttribute =~ /^$notit/)) &&
          		(!($sAttribute =~ /value$/)) &&
          		(!($sAttribute =~ /type$/)) ) {
        		next;
		};

      		# expanding predicate element
      		if( 	($predicateNode) &&
			(!($sAttribute eq $XML::Parser::PenRDF::RDFMS_resource)) ) {
			$expat->xpcroak("Property element ". $ele->{tag}." has invalid attribute ".$sAttribute.". Only rdf:resource is allowed.");
		};

      		$foundAbbreviation = 1;
		my $newElement =  XML::Parser::PenRDF::Element->new($sAttribute);
		my $newData =  XML::Parser::PenRDF::DataElement->new($sValue);
        	push @{$newElement->{children}},$newData;
        	push @{$parent->{children}},$newElement;
    };
    return $foundAbbreviation;
};

sub parseLiteral {
	my ($expat) = @_;

	foreach(reverse @{$expat->{PenRDF}->{elementStack}}) {	
		my $sParseType = getAttributeValue(	$expat,
							$_->{attlist},
							$XML::Parser::PenRDF::RDFMS_parseType );
		if( (defined $sParseType) && ($sParseType ne "Resource") ) {
			return 1;
    		};
	};
    	return 0;       
};

sub parseResource {
	my ($expat) = @_;

	foreach(reverse @{$expat->{PenRDF}->{elementStack}}) {	
		my $sParseType = getAttributeValue(	$expat,
							$_->{attlist},
							$XML::Parser::PenRDF::RDFMS_parseType );
		if( (defined $sParseType) && ($sParseType eq "Resource") ) {
			return 1;
    		};
	};
    	return 0;       
};

sub makeAbsolute {
	my ($expat,$sURI) = @_;

#print STDERR "makeAbsolute(['",$expat->{PenRDF}->{sSource},"'],'$sURI')\n";
	
	my $URL = URI->new($sURI);
        if(defined $URL->scheme) {
                # If sURI is an absolute URI, don't bother
                # with it
		#carp "'$sURI' is already an absolute URI\n";
		return $sURI;
	} elsif(	(defined $sURI) && 
			((defined $expat->{PenRDF}->{sSource}) 
				&& ($expat->{PenRDF}->{sSource} ne '')) ) {
		my $base = URI->new_abs('',$expat->{PenRDF}->{sSource});
		#$sURI ='#'.$sURI # ????
		#	unless($sURI =~ /^#/);
    		my $absoluteURL;
		my $special='';
		if($base->scheme ne 'file') {
			$sURI =~ s/^#//g
				if($base =~ /#$/);
			
			$sURI = uri_escape($sURI);
    			$absoluteURL = URI->new_abs('',$base.$sURI);
		} else {
    			$absoluteURL = $base;
			$special = $sURI;
			$special =~ s/^#//g
				if($base =~ /#$/);
			
			$special = uri_escape($special);
		};
		if(defined $absoluteURL->scheme) {
			return $absoluteURL->as_string.$special;
		} else {
			carp "RDF Resource - cannot combine ", $expat->{PenRDF}->{sSource},
				" with ", $sURI;
	    	};
        } else {
		return $sURI;
	};
};

package XML::Parser::PenRDF::Element;
{
	sub new {
		my ($pkg, $tag, $attlist) = @_;
		$attlist = []
			unless(defined $attlist);
		my $self =  {
				tag		=>	$tag,
				attlist		=>	$attlist,
				sResource	=>	'',
				sAbout		=>	'',
				sAboutEach	=>	'',
				sAboutEachPrefix	=>	'',
				children	=>	[],
				vTargets	=>	[],
				bDone		=>	0,
				sID		=>	'',
				sBagID		=>	'',
				sPrefix		=>	''
			};
		bless $self,$pkg;
	};
};

package XML::Parser::PenRDF::DataElement;
{
	@XML::Parser::PenRDF::DataElement::ISA = qw( XML::Parser::PenRDF::Element );
	sub new {
		my ($pkg, $text, $attlist) = @_;
		my $self = $pkg->SUPER::new($text, $attlist);

		$self->{tag} = "[DATA: " . $text . "]";
		$self->{sContent} = $text; #instanceOf Data :-)
		bless $self,$pkg;
	};
};

1;
};

__END__

=head1 NAME

XML::Parser::PenRDF - This module implements a streaming RDF Parser as a direct implementation of XML::Parser::Expat(3)

=head1 SYNOPSIS

	use XML::Parser::PenRDF;
        use RDF::Store::NodeFactory;
        my $p=new XML::Parser::PenRDF(
		ErrorContext => 2,
                Handlers        => {
                        Init    => sub { print "INIT\n"; },
                        Final   => sub { print "FINAL\n"; },
                        Assert  => sub { print "STATEMENT - @_\n"; }
                },
                NodeFactory     => new RDF::Store::NodeFactory() );

	$p->parsefile('http://www.gils.net/bsr-gils.rdfs');
        $p->parsefile('http://www.gils.net/rdf/bsr-gils.rdfs');
        $p->parsefile('/some/where/my.rdf');
        $p->parsefile('file:/some/where/my.rdf');
	$p->parse(*STDIN);

	use XML::Parser::PenRDF::RDFStore;
	my $pstore=new XML::Parser::PenRDF(
                ErrorContext 	=> 2,
                Style 		=> 'XML::Parser::PenRDF::RDFStore',
                NodeFactory     => new RDF::Store::NodeFactory(),
                Source  	=> 'http://www.gils.net/bsr-gils.rdfs',
                store   =>      {
                                	persistent      =>      1,
                                	directory       =>      '/tmp/',
                                	seevalues       =>      1,
                                	options         =>      { style => 'BerkeleyDB', Q => 20 }
                                }
        );
	$pstore->parsefile('http://www.gils.net/bsr-gils.rdfs');

=head1 DESCRIPTION

This module implements a Resource Description Framework (RDF) I<streaming> parser completely in 
Perl using the XML::Parser::Expat(3) module. The actual RDF parsing happens using an instance of XML::Parser::Expat with Namespaces option enabled and start/stop and char handlers set.
The RDF specific code is based on the modified version of SiRPAC of Sergey Melnik in Java; a lot of
changes and adaptations have been done to actually run it under Perl.
Expat options may be provided when the XML::Parser::PenRDF object is created. These options are then passed on to the Expat object on each parse call.

Exactly like XML::Parser(3) the behavior of the parser is controlled either by the Style entry elsewhere in this document and/or the Handlers entry elsewhere in this document options, or by the setHandlers entry elsewhere in this document method. These all provide mechanisms for XML::Parser::PenRDF to set the handlers needed by Expat.  If neither Style nor Handlers are specified, then parsing just checks the RDF document syntax against the W3C RDF Raccomandation . When underlying handlers get called, they receive as their first parameter the Expat object, not the Parser object.

To see some examples about how to use it look at the sections below and in the samples and utils directory coming with this software distribution.

E.g.
	With XML::Parser::PenRDF you can easily write an rdfingest.pl script to do something like this:

	fetch -o - -q http://dmoz.org/rdf/content.rdf.u8.gz | \
		gunzip - | \
		sed -f dmoz.content.sed | rdfingest.pl - 

=head1 METHODS

=over 4

=item new

This is a class method, the constructor for XML::Parser::PenRDF. Options are passed as keyword value
pairs. Recognized options are:

=over 4

=item * NodeFactory

This option is B<mandatory> to run the XML::Parser::PenRDF parser correctly and must contain a reference to an object of type RDF::NodeFactory(3). Such a reference is used during the RDF parsing to create resources, literal and statements to be passed to the registered handlers. A sample implementation is RDF::Store::NodeFactory that is provided
with the RDFStore package.

=item * Source

This option can be specified by the user to specify a base URI to use for the generation of resource URIs during parsing. If this option is omitted the parser will try to generate a prefix for generated resources using the input filename or URL actually containing the input RDF. In a near future such an option could be obsoleted by use of XMLBase W3C raccomandation.

=item * Style

This option provides an easy way to set a given style of parser. There is one sample Sylte module
provided with the XML::Parser::PenRDF distribution called XML::Parser::PenRDF::RDFStore. Such
a module uses the RDF::Store(3) modules together with the Data::MagicTie(3) to implement a simple
RDF storage.
Custom styles can be provided by giving a full package name containing
at least one '::'. This package should then have subs defined for each
handler it wishes to have installed. See L<"WRITE YOUR OWN PARSER"> below
for a discussion on how to build one.

=item * Handlers

When provided, this option should be an anonymous hash containing as
keys the type of handler and as values a sub reference to handle that
type of event. All the handlers get passed as their 1st parameter the
instance of Expat that is parsing the document. Further details on
handlers can be found in L<"HANDLERS">. Any handler set here
overrides the corresponding handler set with the Style option.

=item * ErrorContext

This is an XML::Parser option. When this option is defined, errors are reported
in context. The value should be the number of lines to show on either side
of the line in which the error occurred.

=back

All the other XML::Parser and XML::Parser::Expat options should work freely with XML::Parser::PenRDF see XML::Parser(3) and XML::Parser::Expat(3).

=back

=item  setHandlers(TYPE, HANDLER [, TYPE, HANDLER [...]])

This method registers handlers for various parser events. It overrides any
previous handlers registered through the Style or Handler options or through
earlier calls to setHandlers. By providing a false or undefined value as
the handler, the existing handler can be unset.

This method returns a list of type, handler pairs corresponding to the
input. The handlers returned are the ones that were in effect prior to
the call.

See a description of the handler types in L<"HANDLERS">.

=item parse(SOURCE [, OPT => OPT_VALUE [...]])

The SOURCE parameter should either be a string containing the whole RDF
document, or it should be an open IO::Handle. Constructor options to
XML::Parser::Expat given as keyword-value pairs may follow the SOURCE
parameter. These override, for this call, any options or attributes passed
through from the XML::Parser::PenRDF instance.

A die call is thrown if a parse error occurs. Otherwise it will return 1
or whatever is returned from the B<Final> handler, if one is installed.
In other words, what parse may return depends on the style.

e.g. the XML::Parser::PenRDF::RDFStore Style module returns an instance of RDF::Model

=item parsestring

This is just an alias for parse for backwards compatibility.

=item parsefile(URL_OR_FILE [, OPT => OPT_VALUE [...]])

Open URL_OR_FILE for reading, then call parse with the open handle. If URL_OR_FILE
is a full qualified URL this module uses chat2.pl to actually fetch the content.

=back

=head1 HANDLERS

As Expat, PenRDF is an event based parser. As the parser recognizes parts of the
RDF document then any handlers registered for that type of an event are called 
with suitable parameters.
All handlers receive an instance of XML::Parser::Expat as their first
argument. See L<XML::Parser::Expat/"METHODS"> for a discussion of the
methods that can be called on this object.

=head2 Init             (Expat)

This is called just before the parsing of the document starts.

=head2 Final            (Expat)

This is called just after parsing has finished, but only if no errors
occurred during the parse. Parse returns what this returns.

=head2 Assert            (Expat, Statement)

This event is generated when a new RDF statement has been generated by the parseer.start tag is recognized. Statement is of type RDF::Statement(3) as generated
by the RDF::NodeFactory(3) passed as argument to the XML::Parser::PenRDF 
constructor.

=head2 Start_XML_Literal            (Expat, Element [, Attr, Val [,...]])

This event is generated when an XML start tag is recognized within an RDF
property with parseType="Literal". Element is the
name of the XML element type that is opened with the start tag. The Attr &
Val pairs are generated for each attribute in the start tag.

This handler should return a string containing either the original XML chunck or one f its transformations, perhaps using XSLT.

=head2 Stop_XML_Literal              (Expat, Element)

This event is generated when an XML end tag is recognized within an RDF
property with parseType="Literal". Note that an XML empty tag (<foo/>) generates both a Start_XML_Literal and an Stop_XML_Literal event.

=head2 Char_XML_Literal             (Expat, String)

This event is generated when non-markup is recognized within an RDF
property with parseType="Literal". The non-markup sequence of characters is in 
String. A single non-markup sequence of encoding of the string in the original 
document, this is given to the handler in UTF-8.

This handler should return the processed text as a string.

=head1 WRITE YOUR OWN PARSER

Write an extension module for you needs it is as easy as write one for XML::Parser :)
Have a look at http://www.xml.com/xml/pub/98/09/xml-perl.html and http://wwwx.netheaven.com/~coopercc/xmlparser/intro.html.

You can either make you Perl script a parser self by embedding the needed function hooks or write a
custom Style module for XML::Parser::PenRDF.

=head2 *.pl scripts

	use XML::Parser::PenRDF;
	use RDF::Store::NodeFactory;
	my $p=new XML::Parser::PenRDF(
		Handlers        => {
			Init    => sub { print "INIT\n"; },
			Final   => sub { print "FINAL\n"; },
			Assert  => sub { print "STATEMENT - @_\n"; }
		},
		NodeFactory     => new RDF::Store::NodeFactory() );


or something like:

	use XML::Parser::PenRDF;
        use RDF::Store::NodeFactory;
	my $p=new XML::Parser::PenRDF( NodeFactory     => new RDF::Store::NodeFactory() );
	$p->setHandlers(        Init    => sub { print "INIT\n"; },
                        	Final   => sub { print "FINAL\n"; },
                        	Assert  => sub { print join(",",@_),"\n"; }     );

=head2 Style modules

A more sophisticated solution is to write a complete Perl5 Sytle module for XML::Parser::PenRDF that
can be easily reused in your code. E.g. a perl script could use this piece of code:

	use XML::Parser::PenRDF;
	use XML::Parser::PenRDF::MyStyle;
	use RDF::Store::NodeFactory;

	my $p=new XML::Parser::PenRDF(	Style => 'XML::Parser::PenRDF::MyStyle',
                			NodeFactory     => new RDF::Store::NodeFactory() );
	$p->parsefile('http://www.gils.net/bsr-gils.rdfs');

The Style module self could stored into a file like MyStyle.pm like this:

	package XML::Parser::PenRDF::MyStyle;

	sub Init { print "INIT\n"; };
	sub Final { print "FINAL\n"; };
	sub Assert {
                print "ASSERT: ",
                                $_[1]->subject()->toString(),
                                $_[1]->predicate()->toString(),
                                $_[1]->object()->toString(), "\n";
	};
	sub Start_XML_Literal { print "STARTAG: ",$_[1],"\n"; };
	sub Stop_XML_Literal { print "ENDTAG: ",$_[1],"\n"; };
	sub Char_XML_Literal { print "UTF8 chrs: ",$_[1],"\n"; };

	1;

For a more complete and useful example see XML::Parser::PenRDF::RDFStore(3).


=head1 BUGS

This module implements most of the W3C RDF Raccomandation as its Java counterpart SiRPAC from the Stanford University Database Group by Sergey Melnik (see http://www-db.stanford.edu/~melnik/rdf/api.html)
This version is conformant to the latest RDF API Draft on 2000-10-27. It does not support yet:

	* aboutEach and aboutEachPrefix

=head1 SEE ALSO

XML::Parser::PenRDF(3), DBMS(3) and XML::Parser(3) XML::Parser::Expat(3)
RDF::Model(3) RDF::Store::NodeFactory(3)

	RDF Model and Syntax Specification - http://www.w3.org/TR/REC-rdf-syntax
	RDF Schema Specification 1.0 - http://www.w3.org/TR/2000/CR-rdf-schema-20000327
	Benchmarking XML Parsers by Clark Cooper - http://www.xml.com/pub/Benchmark/article.html
	See also http://www.w3.org/RDF/Implementations/SiRPAC/SiRPAC-defects.html
	RDF::Parser(3) from http://www.pro-solutions.com

=head1 AUTHOR

	Alberto Reggiori <alberto.reggiori@jrc.it>

	Sergey Melnik <melnik@db.stanford.edu> is the original author of the streaming version of SiRPAC in Java
	Clark Cooper is the author of the XML::Parser(3) module together with Larry Wall
